/*
Protocol analysis based on
"Bounded Cryptographic Protocol Analysis"
ACM CCS-8 [MS01]

[u,v] is concatenation, binary only

u*k is u pk-encrypted with k, usually pk(a)

u/pk(a) is signature of u by a (checksum)

u+k is u encrypted with k as symmetric key

u-k is hidden symmetric encryption

sha is used for a hash function

e is the attacker
*/

/* Some system-specific Prolog is commented out

:- table(solve/2).
:- dynamic(cscount/1).
:- assert(cscount(0)).

% scs resets the constraint set count to zero
% (use it between trials)
% ics increments it by one, used in reach

scs :- 
  retract(cscount(_)),
  assert(cscount(0)).

ics(N1) :-
  retract(cscount(N)),
  N1 is N+1,
  assert(cscount(N1)).
*/

% reach(Bundle,Constraintlist,Terms,FinalConstrtList)
%   Constraintlist initially empty
%     Constraint is [term, termlist]
%   Terms is a list of terms known to attacker
%     Terms initially just principal names
%   Bundle is a list of strands. (Actually a "semibundle")
%     Strand is list of send(M) and recv(M) nodes
% reach merely tests solvability of bundle
% reach creates the initial list of constraints
%   from a possible merge and passes it to solve

reach(B,C,_T,F) :-
  allnull(B),
%  ics(N),
%  write('Try '),write(N),nl,
%  write('Constraints to be solved: '),ppc(C),!,
  solve(C,F).

reach(B,C,T,F) :-
  selectnode(B,send(M),B1),!, % send adds term 
  reach(B1,C,[M|T],F).
  
reach(B,C,T,F) :-
  selectnode(B,recv(M),B1),   % recv adds constraint
  reach(B1,[[M,T]|C],T,F).

% selectnode(B,N,B1) separates B into the first node
% N of some strand and the remaining strand set B1.
% selectnode fails if B is all null.
% Note: to optimize, select all send nodes first, any order.
% (but usually only one send is available anyway)
% If all nodes are recv, order does matter, so all orders
% should be attempted.
% (For debugging purposes, the bundle can be pre-merged
% into a single strand to force an order.)

selectnode(B,send(M),B1) :-
  member([send(M)|_S],B),!,     % this cut for send optimization
  diff(B,send(M),B1).

selectnode([[recv(M)|S]|B],recv(M),[S|B]). % remove recv from first strand,

selectnode([S|B],recv(M),[S|B1]) :-  % or from some other strand
  selectnode(B,recv(M),B1).

% diff(B,N,B1): bundle B minus node N is B1

diff([[N|S]|B],N,[S|B]).

diff([S|B],N,[S|B1]) :- diff(B,N,B1).

% solve([[expr, termlist],...],Varconstraints)
% apply reduct to each nonsimple constraint, in
% reverse (i.e., chronological) order.
% Build up (possibly empty) list of simple [var,termlist] constraints.
% Note that reduct may cause a var to be instantiated
% on the left side of a prior constraint, so
% tail recursion applies solve again from the beginning.

solve(C,C) :- allvarc(C),!.

solve([[A,T]|C],W1) :- 
  solve(C,V),
  remv(T,T1),
  reduct(A,T1,U),
  append(U,V,W),
  solve(W,W1).

% reduct(M,T,C) performs one reduction step on
% an active constraint M:T and 
% records replacement constraints in C

reduct(M,T,[[M,T]]) :-
  var(M),!.		% pass over simple constraint

reduct(M,T,[]) :-  	% (un) with constant
  atomic(M),		% always safe
  member(M,T),!.

reduct([A,B],T,[[B,T],[A,T]]) :- !. 	% (pair), always safe

reduct(M/pk(e),T,[[M,T]]) :- !. 	% (sig), always safe

reduct(M,T,[]) :-		% (un)
  member(A,T),
  hunify(M,A).

reduct(sha(M),T,[[M,T]]).  	% (hash)

reduct(pk(A),T,[[A,T]]).	% public-key lookup

reduct(M*K,T,[[M,T],[K,T]]). 	% (penc)

reduct(M+K,T,[[M,T],[K,T]]).	% (senc)

reduct(M,T,[[M,T]]) :-
  do_ksub(T).

reduct(M,T,[[M,T2],[K,T1]]) :-
  do_ksyn(T,T2,K,T1).

% remv removes variables from a term list, if any.
% It also does (split) and (pdec), they're always safe.

remv([],[]) :- !.

remv([A|T],W) :-
  var(A),!,
  remv(T,W).

remv([[A,B]|T],W) :- !,	% (split)
  remv([A,B|T],W).

remv([U*K|T],W) :-
  K==pk(e),!,		% (pdec)
  remv([U|T],W).

remv([A|T],[A|W]) :-
  remv(T,W).

% do_ksub looks for U*V in a term list
% and binds V to pk(e) if possible (and V not already pk(e))
% It fails if there is no instance to bind.

do_ksub([_U*V|_T]) :-
  \+V==pk(e),
  V=pk(e).

do_ksub([_A|T]) :- do_ksub(T).

% do_ksyn looks for U+K in a term list T
% and decrypts it to U.  We also insert the
% new constraint for K with U-K in T1.
% Fails iff there is no symmetric encryption.

do_ksyn([U+K|T],[U,K|T],K,[U-K|T]).

do_ksyn([A|T],[A|T2],K,[A|T1]) :- 
  do_ksyn(T,T2,K,T1).

% allnull tests for empty bundle

allnull([]).

allnull([[]|B]) :- allnull(B).

% allvarc tests for simple constraint set
%    in which all left sides are variables

allvarc([]).

allvarc([[X,_T]|C]) :-
  var(X),
  allvarc(C).

% hunify(M,A) turns A from - to + first if necessary.

hunify(M,U-V) :- unify(M,U+V),!.
hunify(M,A) :- unify(M,A).

%----------------------------------------------------------------------------
% "safe" unification with occurs check from C. Meadows
%----------------------------------------------------------------------------

unify(X,Y) :-
	var(X),var(Y),!,
	X=Y.
unify(X,Y) :-
	atomic(X),!,X=Y.	% atomic includes numbers
unify(X,Y) :-
	atomic(Y),!,X=Y.
unify(X,Y) :-
	var(X),!,
	notOccurs(X,Y),
	X=Y.
unify(X,Y) :-
	var(Y),!,
	notOccurs(Y,X),
	X=Y.
unify(X,Y) :-
	X =.. [A|B],
	Y =.. [A|C],
	list_unify(B,C),!.

list_unify([],[]).
list_unify([A|B],[C|D]) :-
	unify(A,C),
	list_unify(B,D).

notOccurs(X,Y) :- var(Y),!, \+ X == Y.
notOccurs(_X,Y) :- atomic(Y),!.
notOccurs(X,[Y|Z]) :- !,notOccurs(X,Y),notOccurs(X,Z).
notOccurs(X,Y) :- Y =.. [_F|N],notOccurs(X,N).

 
